#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
from io import StringIO

# Third party modules #

###############################################################################
def r_matrix_to_dataframe(matrix):
    import numpy, pandas
    cols = list(matrix.colnames)
    rows = list(matrix.rownames)
    return pandas.DataFrame(numpy.array(matrix), index=rows, columns=cols)

###############################################################################
def pandas_df_to_r_df(pandas_df):
    import pandas
    from pandas.rpy.common import convert_to_r_dataframe
    return convert_to_r_dataframe(pandas_df)

###############################################################################
def pandas_df_to_named_r_df(pandas_df, r_name):
    eval("%s = convert_to_r_dataframe(pandas_df)" % r_name)
    eval("return %s" % r_name)

###############################################################################
def df_to_pdf(df, pdf_path):
    """Make a PDF from a data frame that is styled with some CSS."""
    # Convert to HTML as a string #
    table = df.to_html(classes='mystyle')
    # Add our custom styling #
    template = '''
    <html>
      <head>
        <title>HTML Pandas Dataframe with CSS</title></head>
        <style>
            .mystyle {
                font-size: 11pt; 
                font-family: Arial;
                border-collapse: collapse; 
                border: 1px solid silver;
            }
            .mystyle td, th {
                padding: 5px;
            }
            .mystyle tr:nth-child(even) {
                background: #E0E0E0;
            }
            .mystyle tr:hover {
                background: silver;
                cursor: pointer;
            }
        </style>
      <body>
        %s
      </body>
    </html>
    '''
    # Use a library #
    import weasyprint
    result = weasyprint.HTML(string=template % table)
    result.write_pdf(pdf_path)

###############################################################################
def string_to_df(string, **kwargs):
    """
    Parse a string as a dataframe. Example:

        >>> from plumbing.dataframes import string_to_df
        >>>
        >>> a = '''  i  | x | y | z
        >>>         AR  | v | 5 | 1
        >>>         For | w | 3 | 3
        >>>         For | w | 4 | 4 '''
        >>>
        >>> df = string_to_df(a)
        >>> print(df)

             i  x  y  z
        0   AR  v  5  1
        1  For  w  3  3
        2  For  w  4  4
    """
    import pandas
    handle = StringIO(string.replace(' ',''))
    df     = pandas.read_csv(handle, sep="|", header=0, **kwargs)
    return df

###############################################################################
def count_unique_index(df, by):
    """
    This function enables you to quickly see how many unique combinations of
    column values exist in a data frame. Here are two examples:

        >>> df = ''' i   | A  | B | C
        >>>          For | 3  | 1 | x
        >>>          For | 3  | 2 | x
        >>>          For | 3  | 3 | y '''
        >>> from plumbing.dataframes import string_to_df
        >>> df = string_to_df(df)
        >>> count_unique_index(df, by=['A', 'C'])

           A  C  count
        0  3  x      2
        1  3  y      1

        >>> import seaborn
        >>> tips = seaborn.load_dataset("tips")
        >>> print(count_unique_index(tips, ['sex', 'smoker']))

              sex smoker  count
        0    Male    Yes     60
        1    Male     No     97
        2  Female    Yes     33
        3  Female     No     54

        >>> tips = seaborn.load_dataset("tips")
        >>> print(count_unique_index(tips, ['sex', 'smoker', 'day']))

              sex smoker    time  count
        0    Male    Yes   Lunch     13
        1    Male    Yes  Dinner     47
        2    Male     No   Lunch     20
        3    Male     No  Dinner     77
        4  Female    Yes   Lunch     10
        5  Female    Yes  Dinner     23
        6  Female     No   Lunch     25
        7  Female     No  Dinner     29
    """
    return df.groupby(by).size().reset_index().rename(columns={0: 'count'})

###############################################################################
def multiply_propagate_nan(df, series):
    """
    This function enables you to multiply a dataframe by a series with 'smart'
    propagation of NaN values.

    We implement a simple rule where:

        x * NaN = NaN | ∀x≠0
        0 * NaN = 0

    In essence, multipling a NaN with a zero will yield a zero and multipling
    a NaN with anything else will yield a NaN.
    This is usefull in some scenarios where NaN simply indicates a missing
    value in a dataset, which is known to be a dataset of real numbers.
    For instance, when a scientist knows that all values considered in a given
    experiment are scalars but a few of the values were not recorded or were
    lost.
    This is done because, as it stands, NaN can represent anything even the
    result of an operation yielding infinity. But there is no float named "IaN"
    that would stand for "Is a Number", where we know that the number is
    in R or N, but are missing its particular value.

    #TODO restore zeros from the series too not only the dataframe.

    An example follows:

        >>> one = '''    | W1   | W2   | W3
        >>>           A  | 0.0  | 0.7  | NaN
        >>>           B  | 0.1  | 1.0  | 1.0
        >>>           C  | 0.2  | 0.8  | 0.49
        >>>           D  | 0.0  | 0.0  | 1.0  '''
        >>>
        >>> two = '''    | Y
        >>>           W3 | 8.0
        >>>           W1 | NaN
        >>>           W2 | 9.0  '''
        >>>
        >>> from plumbing.dataframes import string_to_df
        >>> one = string_to_df(one, index_col=0)
        >>> two = string_to_df(two, index_col=0)
        >>>
        >>> from plumbing.dataframes import multiply_propagate_nan
        >>> print(multiply_propagate_nan(one, two['Y']))

                W1   W2    W3
            A  0.0  6.3   NaN
            B  NaN  9.0  8.00
            C  NaN  7.2  3.92
            D  0.0  0.0  8.00
    """
    # Multiply #
    result = df * series
    # Restore the zeros where they should be #
    result[df == 0.0] = 0.0
    # Return #
    return result


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

"""
Typically you would run this file from a command line like this:

    ipython.exe -i -- /deploy/plumbing/tests/database/access_db/test_copy_table.py

Or in the Ubuntu WSL:

    ipython -i -- ~/deploy/plumbing/tests/database/access_db/test_copy_table.py
"""

# Built-in module #
import inspect, os

# Internal modules #
from plumbing.databases.access_database import AccessDatabase
from autopaths.file_path import FilePath

# Third party modules #

# Constants #
file_name = inspect.getframeinfo(inspect.currentframe()).filename
this_dir  = os.path.dirname(os.path.abspath(file_name)) + '/'

# Never modify the original #
orig_path    = FilePath(this_dir + 'orig.mdb')
testing_path = FilePath(this_dir + 'testing.mdb')
orig_path.copy(testing_path)

# The source database #
source_db = AccessDatabase(testing_path)

# The destination database #
dest_path = FilePath(this_dir + 'copied_table.mdb')
dest_db   = AccessDatabase.create(dest_path)

# Copy a table #
dest_db.import_table(source_db, "tblClassifierSets")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import sys, os, time, re, random, math, json
import getpass, hashlib, collections
import unicodedata
from itertools import compress, product

# One-liners #
flatter = lambda x: [item for sublist in x for item in sublist]

###############################################################################
def rich_panel_print(text, title=None):
    """Make a pretty box around a text with the `rich` library."""
    from rich import print as rich_print
    from rich.panel import Panel
    from rich.padding import Padding
    rich_print(Padding(Panel(text, title=title, padding=2, expand=False), (2,10)))

###############################################################################
def no_single_newline(text):
    """
    Remove single newlines in a string, but preserve multiple newlines.
    Taken from https://stackoverflow.com/questions/22649851
    """
    return re.sub('(?<![\r\n])(\r?\n|\n?\r)(?![\r\n])', ' ', text)

###############################################################################
def ascii(text):
    """
    Make a safe, ASCII version of a string. For instance, for use on the web.
    """
    return unicodedata.normalize('NFKD', unicode(text)).encode('ASCII', 'ignore')

def alphanumeric(text):
    r"""
    Make an ultra-safe, ASCII version a string.
    For instance for use as a filename.
    \w matches any alphanumeric character and the underscore.
    """
    return "".join([c for c in text if re.match(r'\w', c)])

###############################################################################
def sanitize_text(text):
    r"""
    Make a safe representation of a string.
    Note: the `\s` special character matches any whitespace character.
    This is equivalent to the set [\t\n\r\f\v] as well as ` ` (whitespace).
    """
    # First replace characters that have specific effects with their repr #
    text = re.sub("(\\s)", lambda m: repr(m.group(0)).strip("'"), text)
    # Make it a unicode string (the try supports python 2 and 3) #
    try: text = text.decode('utf-8')
    except AttributeError: pass
    # Normalize it “
    text = unicodedata.normalize('NFC', text)
    return text

###############################################################################
def camel_to_snake(text):
    """
    Will convert CamelCaseStrings to snake_case_strings. Examples:

    >>> camel_to_snake('CamelCase')
    'camel_case'
    >>> camel_to_snake('CamelCamelCase')
    'camel_camel_case'
    >>> camel_to_snake('Camel2Camel2Case')
    'camel2_camel2_case'
    >>> camel_to_snake('getHTTPResponseCode')
    'get_http_response_code'
    >>> camel_to_snake('get2HTTPResponseCode')
    'get2_http_response_code'
    >>> camel_to_snake('HTTPResponseCode')
    'http_response_code'
    >>> camel_to_snake('HTTPResponseCodeXYZ')
    'http_response_code_xyz'
    >>> camel_to_snake('Double_Case')
    'double_case'
    >>> camel_to_snake('SW_Merch')
    'sw_merch'
    >>> camel_to_snake('Odd/Characters')
    'odd_characters'
    >>> camel_to_snake('With Spaces')
    'with_spaces'
    """
    # Load #
    result = text
    # Eliminate trailing spaces #
    result = result.strip(' ')
    # First step #
    try:
        result = re.sub('(.)([A-Z][a-z]+)',  r'\1_\2', result)
    except TypeError:
        print("The text received was '%s'" % result)
        raise
    # Second step #
    result = re.sub('([a-z0-9])([A-Z])', r'\1_\2', result)
    # Lower case the rest #
    result = result.lower()
    # Eliminate remaining spaces #
    result = result.replace(' ', '_')
    # Eliminate quote characters #
    result = result.replace('"', '')
    result = result.replace("'", '')
    # Eliminate parenthesis #
    result = result.replace("(", '')
    result = result.replace(")", '')
    # Eliminate special characters #
    result = result.replace('/', '_')
    # Eliminate double underscore #
    while '__' in result: result = result.replace('__', '_')
    # Return #
    return result

###############################################################################
def bool_to_unicode(b):
    """
    Different possibilities for True: ☑️✔︎✓✅👍✔️
    Different possibilities for False: ✕✖︎✗✘✖️❌⛔️❎👎🛑🔴
    """
    b = bool(b)
    if b is True:  return u"✅"
    if b is False: return u"🔴"

###############################################################################
def access_dict_like_obj(obj, prop, new_value=None):
    """
    Access a dictionary like if it was an object with properties.
    If no "new_value", then it's a getter, otherwise it's a setter.
    >>> {'characters': {'cast': 'Jean-Luc Picard', 'featuring': 'Deanna Troi'}}
    >>> access_dict_like_obj(startrek, 'characters.cast', 'Pierce Brosnan')
    """
    props = prop.split('.')
    if new_value:
        if props[0] not in obj: obj[props[0]] = {}
        if len(props)==1: obj[prop] = new_value
        else: return access_dict_like_obj(obj[props[0]], '.'.join(props[1:]), new_value)
    else:
        if len(props)==1: return obj[prop]
        else: return access_dict_like_obj(obj[props[0]], '.'.join(props[1:]))

###############################################################################
def all_combinations(items):
    """Generate all combinations of a given list of items."""
    return (set(compress(items,mask)) for mask in product(*[[0,1]]*len(items)))

###############################################################################
def pad_equal_whitespace(string, pad=None):
    """Given a multiline string, add whitespaces to every line
    so that every line has the same length."""
    if pad is None: pad = max(map(len, string.split('\n'))) + 1
    return '\n'.join(('{0: <%i}' % pad).format(line) for line in string.split('\n'))

###############################################################################
def pad_extra_whitespace(string, pad):
    """Given a multiline string, add extra whitespaces to the front of every line."""
    return '\n'.join(' ' * pad + line for line in string.split('\n'))

###############################################################################
def mirror_lines(string):
    """Given a multiline string, return its reflection along a vertical axis.
    Can be useful for the visualization of text version of trees."""
    return '\n'.join(line[::-1] for line in string.split('\n'))

###############################################################################
def concatenate_by_line(first, second):
    """Zip two strings together, line wise"""
    return '\n'.join(x+y for x,y in zip(first.split('\n'), second.split('\n')))

###############################################################################
def sort_string_by_pairs(strings):
    """Group a list of strings by pairs, by matching those with only
    one character difference between each other together."""
    assert len(strings) % 2 == 0
    pairs = []
    strings = list(strings) # This shallow copies the list
    while strings:
        template = strings.pop()
        for i, candidate in enumerate(strings):
            if count_string_diff(template, candidate) == 1:
                pair = [template, strings.pop(i)]
                pair.sort()
                pairs.append(pair)
                break
    return pairs

###############################################################################
def count_string_diff(a,b):
    """Return the number of characters in two strings that don't exactly match"""
    shortest = min(len(a), len(b))
    return sum(a[i] != b[i] for i in range(shortest))

###############################################################################
def iflatten(L):
    """Iterative flatten."""
    for sublist in L:
        if hasattr(sublist, '__iter__'):
            for item in iflatten(sublist): yield item
        else: yield sublist

###############################################################################
def uniquify_list(L):
    """Same order unique list using only a list compression."""
    return [e for i, e in enumerate(L) if L.index(e) == i]

###############################################################################
def average(iterator):
    """Iterative mean."""
    count = 0
    total = 0
    for num in iterator:
        count += 1
        total += num
    return float(total)/count

###############################################################################
def round_to_halves(number):
    """Round a number to the closest half integer.
    >>> round_to_halves(1.3)
    1.5
    >>> round_to_halves(2.6)
    2.5
    >>> round_to_halves(3.0)
    3.0
    >>> round_to_halves(4.1)
    4.0
    """
    return round(number * 2) / 2

###############################################################################
def get_next_item(iterable):
    """Gets the next item of an iterable.
    If the iterable is exhausted, returns None."""
    try: x = iterable.next()
    except StopIteration: x = None
    except AttributeError: x = None
    return x

###############################################################################
def pretty_now():
    """Returns something like '2019-02-15 15:58:22 CET+0100'"""
    import datetime, tzlocal
    time_zone = tzlocal.get_localzone()
    now       = datetime.datetime.now(time_zone)
    return now.strftime("%Y-%m-%d %H:%M:%S %Z%z")

###############################################################################
def andify(list_of_strings):
    """
    Given a list of strings will join them with commas
    and a final "and" word.

    >>> andify(['Apples', 'Oranges', 'Mangos'])
    'Apples, Oranges and Mangos'
    """
    result = ', '.join(list_of_strings)
    comma_index = result.rfind(',')
    if comma_index > -1: result = result[:comma_index] + ' and' + result[comma_index+1:]
    return result

###############################################################################
def num_to_ith(num):
    """1 becomes 1st, 2 becomes 2nd, etc."""
    value             = str(num)
    before_last_digit = value[-2]
    last_digit        = value[-1]
    if len(value) > 1 and before_last_digit == '1': return value +'th'
    if last_digit == '1': return value + 'st'
    if last_digit == '2': return value + 'nd'
    if last_digit == '3': return value + 'rd'
    return value + 'th'

def isubsample(full_sample, k, full_sample_len=None):
    """Down-sample an enumerable list of things"""
    # Determine length #
    if not full_sample_len: full_sample_len = len(full_sample)
    # Check size coherence #
    if not 0 <= k <= full_sample_len:
        raise ValueError('Required that 0 <= k <= full_sample_length')
    # Do it #
    picked = 0
    for i, element in enumerate(full_sample):
        prob = (k - picked) / (full_sample_len - i)
        if random.random() < prob:
            yield element
            picked += 1
    # Did we pick the right amount #
    assert picked == k

###############################################################################
def sum_vectors_with_padding(vectors):
    """
    Given an arbitrary amount of NumPy one-dimensional vectors of floats,
    do an element-wise sum, padding with 0 any that are shorter than the
    longest array (see https://stackoverflow.com/questions/56166217).

    >>> v1 = numpy.array([0, 0, 5, 5, 1, 1, 1, 1, 0, 0])
    >>> v2 = numpy.array([4, 4, 4, 5, 5, 0, 0])
    >>> v3 = numpy.array([1, 1, 1])
    >>> sum_vectors_with_padding([v1, v2, v3])
    array([ 5,  5, 10, 10,  6,  1,  1,  1,  0,  0])
    """
    import numpy
    all_lengths = [len(i) for i in vectors]
    max_length  = max(all_lengths)
    out         = numpy.zeros(max_length)
    for l,v in zip(all_lengths, vectors): out[:l] += v
    return out

###############################################################################
def moving_average(interval, windowsize, borders=None):
    """
    This is essentially a convolution operation
    Several option exist for dealing with the border cases.

        * None: Here the returned signal will be smaller than the inputted interval.

        * zero_padding: Here the returned signal will be larger than the inputted
        interval and we will add zeros to the original interval before operating
         the convolution.

        * zero_padding_and_cut: Same as above only the result is truncated to be
         the same size as the original input.

        * copy_padding: Here the returned signal will be larger than the inputted
         interval and we will use the right and leftmost values for padding before
          operating the convolution.

        * copy_padding_and_cut: Same as above only the result is truncated to be
         the same size as the original input.

        * zero_stretching: Here we will compute the convolution only in the valid domain,
         then add zeros to the result so that the output is the same size as the input.

        * copy_stretching: Here we will compute the convolution only in the valid domain,
         then copy the right and leftmost values so that the output is the same
          size as the input.
        """
    # The window size in half #
    half = int(math.floor(windowsize/2.0))
    # The normalized rectangular signal #
    import numpy
    window = numpy.ones(int(windowsize))/float(windowsize)
    # How do we deal with borders #
    if borders is None:
        return numpy.convolve(interval, window, 'valid')
    if borders == 'zero_padding':
        return numpy.convolve(interval, window, 'full')
    if borders == 'zero_padding_and_cut':
        return numpy.convolve(interval, window, 'same')
    if borders == 'copy_padding':
        new_interval = [interval[0]]*(windowsize-1) + interval + [interval[-1]]*(windowsize-1)
        return numpy.convolve(new_interval, window, 'valid')
    if borders == 'copy_padding_and_cut':
        new_interval = [interval[0]]*(windowsize-1) + interval + [interval[-1]]*(windowsize-1)
        return numpy.convolve(new_interval, window, 'valid')[half:-half]
    if borders == 'zero_stretching':
        result = numpy.convolve(interval, window, 'valid')
        pad = numpy.zeros(half)
        return numpy.concatenate((pad, result, pad))
    if borders == 'copy_stretching':
        result = numpy.convolve(interval, window, 'valid')
        left = numpy.ones(half)*result[0]
        right = numpy.ones(half)*result[-1]
        return numpy.concatenate((left, result, right))

###############################################################################
def wait(predicate, interval=1, message=lambda: "Waiting..."):
    """Wait until the predicate turns true and display a turning ball."""
    ball, next_ball = u"|/-\\", "|"
    sys.stdout.write("    \033[K")
    sys.stdout.flush()
    while not predicate():
        time.sleep(interval)
        next_ball = ball[(ball.index(next_ball) + 1) % len(ball)]
        sys.stdout.write("\r " + str(message()) + " " + next_ball + " \033[K")
        sys.stdout.flush()
    print("\r Done. \033[K")
    sys.stdout.flush()

###############################################################################
def natural_sort(item):
    """
    Sort strings that contain numbers correctly. Works in Python 2 and 3.

    >>> l = ['v1.3.12', 'v1.3.3', 'v1.2.5', 'v1.2.15', 'v1.2.3', 'v1.2.1']
    >>> l.sort(key=natural_sort)
    >>> l.__repr__()
    "['v1.2.1', 'v1.2.3', 'v1.2.5', 'v1.2.15', 'v1.3.3', 'v1.3.12']"
    """
    dre = re.compile(r'(\d+)')
    return [int(s) if s.isdigit() else s.lower() for s in re.split(dre, item)]

###############################################################################
def split_thousands(s):
    """
    Splits a number on thousands.

    >>> split_thousands(1000012)
    "1'000'012"
    """
    # Check input #
    if s is None: return "0"
    # If it's a string #
    if isinstance(s, str): s = float(s)
    # If it's a float that should be an int #
    if isinstance(s, float) and s.is_integer(): s = int(s)
    # Use python built-in #
    result = "{:,}".format(s)
    # But we want single quotes #
    result = result.replace(',', "'")
    # Return #
    return result

###############################################################################
def gps_deg_to_float(data):
    m = re.search("(\\d+?)°(\\d+?)\'(\\d+?)\'\'", data.strip())
    degs, mins, secs = [0.0 if m.group(i) is None else int(m.group(i)) for i in range(1, 4)]
    comp_dir = -1 if data[-1] in ('N', 'E') else 1
    return (degs + (mins / 60) + (secs / 3600)) * comp_dir

###############################################################################
def is_integer(string):
    try: int(string)
    except ValueError: return False
    return True

###############################################################################
def reverse_compl_with_name(old_seq):
    """Reverse a SeqIO sequence, but keep its name intact."""
    new_seq = old_seq.reverse_complement()
    new_seq.id = old_seq.id
    new_seq.description = old_seq.description
    return new_seq

###############################################################################
class GenWithLength(object):
    """A generator with a length attribute"""
    def __init__(self, gen, length): self.gen, self.length = gen, length
    def __iter__(self): return self.gen
    def __len__(self):  return self.length

###############################################################################
class Password(object):
    """A password object that will only prompt the user once per session."""
    def __str__(self): return self.value
    def __init__(self, prompt=None):
        self._value = None
        self.prompt = prompt

    @property
    def value(self):
        if self._value is None: self._value = getpass.getpass(self.prompt)
        return self._value

###############################################################################
def load_json_path(path):
    """Load a file with the json module, but report errors better if it
    fails. And have it ordered too !"""
    with open(path) as handle:
        try: return json.load(handle, object_pairs_hook=collections.OrderedDict)
        except ValueError as error:
            message = "Could not decode JSON file '%s'." % path
            message = "-"*20 + "\n" + message + "\n" + str(error) + "\n" + "-"*20 + "\n"
            sys.stderr.write(message)
            raise error

###############################################################################
def find_file_by_name(name, root=os.curdir):
    for dirpath, dirnames, filenames in os.walk(os.path.abspath(root)):
        if name in filenames: return os.path.join(dirpath, name)
    raise Exception("Could not find file '%s' in '%s'" % (name, root))

###############################################################################
def md5sum(file_path, blocksize=65536):
    """Compute the md5 of a file. Pretty fast."""
    md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for block in iter(lambda: f.read(blocksize), ""):
            md5.update(block)
    return md5.hexdigest()

###############################################################################
def reversed_lines(path):
    """Generate the lines of file in reverse order."""
    with open(path, 'r') as handle:
        part = ''
        for block in reversed_blocks(handle):
            for c in reversed(block):
                if c == '\n' and part:
                    yield part[::-1]
                    part = ''
                part += c
        if part: yield part[::-1]

###############################################################################
def reversed_blocks(handle, blocksize=4096):
    """Generate blocks of file's contents in reverse order."""
    handle.seek(0, os.SEEK_END)
    here = handle.tell()
    while 0 < here:
        delta = min(blocksize, here)
        here -= delta
        handle.seek(here, os.SEEK_SET)
        yield handle.read(delta)

###############################################################################
def prepend_to_file(path, data, bufsize=1<<15):
    """TODO:
    * Add a random string to the backup file.
    * Restore permissions after copy.
    """
    # Backup the file #
    backupname = path + os.extsep + 'bak'
    # Remove previous backup if it exists #
    try: os.unlink(backupname)
    except OSError: pass
    os.rename(path, backupname)
    # Open input/output files,  note: outputfile's permissions lost #
    with open(backupname) as inputfile:
        with open(path, 'w') as outputfile:
            outputfile.write(data)
            buf = inputfile.read(bufsize)
            while buf:
                outputfile.write(buf)
                buf = inputfile.read(bufsize)
    # Remove backup on success #
    os.remove(backupname)

def append_to_file(path, data):
    with open(path, "a") as handle:
        handle.write(data)

###############################################################################
def tail(path, window=20):
    with open(path, 'r') as f:
        BUFSIZ = 1024
        f.seek(0, 2)
        num_bytes = f.tell()
        size = window + 1
        block = -1
        data = []
        while size > 0 and num_bytes > 0:
            if num_bytes - BUFSIZ > 0:
                # Seek back one whole BUFSIZ
                f.seek(block * BUFSIZ, 2)
                # Read BUFFER
                data.insert(0, f.read(BUFSIZ))
            else:
                # File too small, start from beginning
                f.seek(0,0)
                # Only read what was not read
                data.insert(0, f.read(num_bytes))
            lines_found = data[0].count('\n')
            size -= lines_found
            num_bytes -= BUFSIZ
            block -= 1
        return '\n'.join(''.join(data).splitlines()[-window:])

def head(path, lines=20):
    with open(path, 'r') as handle:
        return ''.join(handle.next() for line in xrange(lines))

###############################################################################
def which(cmd_name, safe=False):
    """
    Portable Python implementation of the unix `which` command.
    Will return the full path of a given executable if it is found
    somewhere in the `$PATH`.
    Otherwise we return `None` or alternatively we raise an exception when
    `safe` is set to False.
    Note: Python 3.3 and above also offers `shutil.which()`.
    Based on the package at: https://github.com/jc0n/python-which
    And the answers at: https://stackoverflow.com/questions/377017
    """
    # The name of the command cannot contain separators #
    assert os.sep not in cmd_name
    # We will use our own first party package #
    from autopaths.file_path import FilePath
    # Loop over every path in the environment variable #
    for path in os.environ['PATH'].split(os.pathsep):
        path = os.path.expanduser(path)
        candidate = os.path.join(path, cmd_name)
        if not os.path.isfile(candidate):     continue
        if not os.access(candidate, os.X_OK): continue
        return FilePath(candidate)
    # Raise an exception optionally #
    msg = 'which() failed to locate a proper command path for "%s"' % cmd_name
    if not safe: raise Exception(msg)

###############################################################################
def query_yes_no(question, default="no"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:    prompt = " [y/n] "
    elif default == "yes": prompt = " [Y/n] "
    elif default == "no":  prompt = " [y/N] "
    else: raise ValueError("invalid default answer: '%s'" % default)
    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '': return valid[default]
        elif choice in valid:                    return valid[choice]
        else: sys.stdout.write("Please respond with 'yes' or 'no'\n")

###############################################################################
def pretty_print_xml(xml_string=None, xml_path=None):
    """
    Will convert XML to a more human readable version.
    You should either pass an XML string, or a path to an XML file.
    In the latter, we will convert it in place, and rewrite the file.
    """
    # Check if we got a file #
    if xml_path is not None:
        with open(xml_path) as handle:
            xml_string = handle.read()
    # Convert it #
    import xml.etree.ElementTree as etree
    element = etree.XML(xml_string)
    etree.indent(element)
    xml_string = etree.tostring(element, encoding='unicode')
    # Check if we got a file #
    if xml_path is not None:
        with open(xml_path, 'w') as handle:
            handle.write(xml_string)

###############################################################################
def combine_pdf_pages(input_pdfs, out_path):
    """
    Given a list of PDF files, we will combine all the pages together in a
    single output file.
    """
    import PyPDF2
    merger = PyPDF2.PdfFileMerger()
    for pdf in input_pdfs:
        merger.append(pdf)
    merger.write(out_path)
    merger.close()

##############################################################################
class LazyStringIO:
    """
    A class that mimics the behavior of StringIO but integrates an iterator.
    This version uses an internal buffer to store data instead of StringIO.

    Usage:

        >>> iterator = iter([b"Hello", b", ", b"World", b"!"])
        >>>
        >>> with LazyStringIO(iterator) as stream:
        >>>     print(stream.read(7))    # Hello//,
        >>>     print(stream.read(8))    # //World//
        >>>     print(stream.read())     # !
    """

    def __init__(self, iterator, separator="//\n", encoding="latin1"):
        self.buffer    = ""
        self.iterator  = iterator
        self.separator = separator
        self.encoding  = encoding
        self.exhausted = False

    def fetch_next(self):
        try:
            return next(self.iterator).decode(self.encoding)
        except StopIteration:
            self.exhausted = True
            return None

    def read(self, n=-1):
        """If n is -1 (default value), it means read all"""
        # Use the buffer #
        values = [self.buffer]
        total  = len(self.buffer)
        self.buffer = ''
        # Now we will read just enough items #
        while True:
            # If we have enough data #
            if 0 < n <= total: break
            # Get one value and add it #
            next_value = self.fetch_next()
            # If we reached the end #
            if next_value is None: break
            # Add it with a separator #
            values += [next_value, self.separator]
            # Calculate the amount of data we have #
            total += len(next_value) + len(self.separator)
        # Compute the long string #
        data = ''.join(values)
        # If we still didn't get enough, send everything we have #
        if n < 0 or total <= n: return data
        # We have more than enough #
        result = data[:n]
        self.buffer = data[n:]
        # Return #
        return result

    def close(self):
        pass  # Nothing to do for closing in this refactored version

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Constants #
popular_agents = None

###############################################################################
def make_headers(user_agent = None):
    """
    Will determine the right headers to pass along with the request.

    If you set the user_agent to None, we will send the default one
    which looks like: "python-requests/2.2.1 CPython/2.7.5 Darwin/13.1.0"
    If you set the user agent to a string, we will send that string.
    If you set the user agent to an integer N we will attempt to use the
    Nth most common user agent string currently in use on the web.
    It will look something like: "'Mozilla/5.0 (Windows NT 10.0; Win64; x64)
    AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36'"
    """
    # We start with nothing #
    headers = {}
    # If we got a string for user agent #
    if isinstance(user_agent, str):
        headers.update({'User-Agent': user_agent})
    # If we got a number for user agent #
    if isinstance(user_agent, int):
        # Load the file from disk if it's not already #
        global popular_agents
        if popular_agents is None:
            from .run_get_user_agents import destin
            popular_agents = destin.lines
        # Remove forbidden characters #
        string = popular_agents[user_agent-1]
        string = string.strip('\n')
        # Update #
        headers.update({'User-Agent': string})
    # Return #
    return headers


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in module #
import inspect, os

# Internal modules #
from plumbing.redirect import RedirectStdStreams

# Constants #
file_name = inspect.getframeinfo(inspect.currentframe()).filename
this_dir  = os.path.dirname(os.path.abspath(file_name)) + '/'

# Test #
print('I should be in the standard out')
with open(this_dir + 'test_redirect.log', 'w') as log_handle:
    with RedirectStdStreams(stdout=log_handle, stderr=log_handle):
        print("I should go to a file")
print("I should also be in the standard out")


#!/usr/bin/env python2
# -*- coding: utf-8 -*-

"""
Typically you would run this file from a command line like this:

     ipython.exe -i -- /deploy/plumbing/tests/database/access_db/time_conversion.py

Or in the Ubuntu WSL:

    ipython -i -- ~/deploy/plumbing/tests/database/access_db/time_conversion.py
"""

# Built-in module #
import inspect, os

# Internal modules #
from plumbing.databases.access_database import AccessDatabase
from plumbing.timer import Timer

# First party modules #
from autopaths.file_path import FilePath

# Constants #
file_name = inspect.getframeinfo(inspect.currentframe()).filename
this_dir  = os.path.dirname(os.path.abspath(file_name)) + '/'

# Paths #
big_mdb_path = FilePath(this_dir + 'big.mdb')

# Create example db #
from cbm_explorer.continent import continent
original = continent.first.run_database
original.copy(big_mdb_path)

# The database #
big_mdb = AccessDatabase(big_mdb_path)

# Clean up #
big_sql = big_mdb.replace_extension('sqlite')
big_sql.remove()

# Convert #
#with Timer(): big_mdb.convert_to_sqlite(method="shell", progress=True)
#with Timer(): big_mdb.convert_to_sqlite(method="dataframe", progress=True)
#with Timer(): big_mdb.convert_to_sqlite(method="object", progress=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import sys

###############################################################################
class RedirectStdStreams(object):
    """
    See file `plumbing/test/redirect/test_redirect.py`
    """

    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    def __enter__(self):
        # Save old ones #
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        # Flush everything #
        self.old_stdout.flush()
        self.old_stderr.flush()
        # Redirect #
        sys.stdout = self._stdout
        sys.stderr = self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        # Flush #
        self._stdout.flush()
        self._stderr.flush()
        # Restore #
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

###############################################################################
class SuppressAllOutput(object):
    """For those annoying modules that can't shut-up about warnings."""

    def __enter__(self):
        # Standard error #
        sys.stderr.flush()
        self.old_stderr = sys.stderr
        sys.stderr = open('/dev/null', 'a+', 0)
        # Standard out #
        sys.stdout.flush()
        self.old_stdout = sys.stdout
        sys.stdout = open('/dev/null', 'a+', 0)

    def __exit__(self, exc_type, exc_value, traceback):
        # Standard error #
        sys.stderr.flush()
        sys.stderr = self.old_stderr
        # Standard out #
        sys.stdout.flush()
        sys.stdout = self.old_stdout

    def example(self):
        print("printing to stdout before suppression", file=sys.stdout, flush=True)
        print("printing to stderr before suppression", file=sys.stderr, flush=True)
        with SuppressAllOutput():
            print("printing to stdout during suppression", file=sys.stdout, flush=True)
            print("printing to stderr during suppression", file=sys.stderr, flush=True)
        print("printing to stdout after suppression", file=sys.stdout, flush=True)
        print("printing to stderr after suppression", file=sys.stderr, flush=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import re

# Modules that's don't have the same PyPI name as import name #
module_names = {
    "biopython":    "Bio",
    "ipython":      "IPython",
    "scikit-learn": "sklearn",
    "scikit-bio":   "skbio",
    "biom-format":  "biom",
}

###############################################################################
def check_setup_py(path_of_setup):
    """
    Parses the required modules from a `setup.py` file and checks they are
    importable and have the minimum required version installed.

    Some ideas for extracting dependency information from a `setup.py` file:
    https://stackoverflow.com/questions/24236266/
    https://stackoverflow.com/questions/27790297/

    Another option is to try the `prasesetup` package, but it is buggy
    and won't even install via pip:
    https://github.com/benfred/parsesetup/issues/3

    Note: In this method, the code in the setup.py will be evaluated.

    Other interesting projects:
    https://pypi.org/project/requirements-parser/

    Typically you can use this function like this:

        >>> from plumbing.dependencies import check_setup_py
        >>> check_setup_py('~/module_name/setup.py')
    """
    # Import packages #
    from unittest import mock
    import setuptools
    from runpy import run_path
    # Parse it #
    from autopaths.file_path import FilePath
    path_of_setup = FilePath(path_of_setup)
    # Run the setup in a mock #
    with mock.patch.object(setuptools, 'setup') as mock_setup:
        run_path(path_of_setup)
    # Retrieve arguments that were used in the call #
    args, kwargs = mock_setup.call_args
    requires     = kwargs['install_requires']
    # Parse each requirement separately #
    requires = [re.split(r'==|>=', req) for req in requires]
    requires = [req if len(req) == 2 else (req[0], None) for req in requires]
    requires = dict(requires)
    # Loop #
    for package, version in requires.items(): check_module(package, version)

###############################################################################
def check_module(mod_name, min_version=None):
    """
    Raise an exception if the given module is either not found or has a
    version that is below the minimum required version number (if specified).

    Typically you can use this function like this:

        >>> from plumbing.dependencies import check_module
        >>> check_module('lxml', '4.5.3')

    If this is not the case the following exception will be raised:

        ImportError: The minimum version required for the "lxml" package is
        <4.5.3>. You have version <4.5.2> installed. Please update the package
        to the latest version and try again.
    """
    # Special cases with modules that change names #
    if mod_name in module_names:
        pypi_name = mod_name
        imrt_name = module_names[mod_name]
        full_name = pypi_name + ' (' + imrt_name + ')'
    else:
        pypi_name = mod_name
        imrt_name = mod_name
        full_name = mod_name
    # Special case for `sh` which doens't set __spec__ #
    if imrt_name == 'sh': return
    # Use a function available in python >= 3.6 #
    import importlib
    spec = importlib.util.find_spec(imrt_name)
    # Message #
    msg = 'You do not seem to have the "%s" package installed. Either you' \
          ' never installed it or your `$PYTHONPATH` is not set up' \
          ' correctly. For more instructions see the `README.md` file.'
    # If we didn't find it #
    if spec is None: raise ModuleNotFoundError(msg % full_name)
    # Did the user request we also check the version #
    if min_version is None: return
    # Just import it and capture the version #
    module = __import__(imrt_name)
    if hasattr(module, '__version__'):
        cur_version = module.__version__
    else:
        # Otherwise use a function available in python >= 3.8 #
        from importlib.metadata import version
        cur_version = version(imrt_name)
    # Message #
    msg = 'The minimum version required for the "%s" package is <%s>.' \
          ' You have version <%s> installed. Please update the package to' \
          ' the latest version and try again.'
    msg = msg % (full_name, min_version, cur_version)
    # Parse the versions #
    from packaging.version import Version
    cur_version = Version(cur_version)
    min_version = Version(min_version)
    # If the version is too low #
    if cur_version < min_version: raise ImportError(msg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os

# Internal modules #
from plumbing.scraping.headers import make_headers

# First party modules #
import autopaths
from autopaths import Path

# Third party modules #
import requests
from retry import retry

###############################################################################
@retry(requests.exceptions.HTTPError, tries=8, delay=1, backoff=2)
def request(url,
            header   = None,
            text     = False,
            content  = False,
            response = False,
            **kwargs):
    # Get #
    resp = requests.get(url, headers=header, **kwargs)
    # This will be caught by the decorator #
    resp.raise_for_status()
    # Pick what to return #
    if text:     return resp.text
    if content:  return resp.content
    if response: return resp

###############################################################################
def retrieve_from_url(url, user_agent=None, **kwargs):
    """
    Return the text content of a resource (e.g. the HTML).
    By default, we will retry if an HTTP error arises.
    """
    # Custom user agent if needed #
    if user_agent is not None: header = make_headers(user_agent)
    else: header = ""
    # Download #
    content = request(url, header, text=True, **kwargs)
    # Return #
    return content

###############################################################################
def stream_from_url(*args, **kwargs):
    """
    Save the resource as a file on disk iteratively by first asking
    for the 'content-length' header entry and downloading in chunks.
    By default, we will retry if an HTTP error arises.
    By default, we will uncompress a downloaded file if it is zipped.
    """
    # Just redirect to download_from_url #
    kwargs.update({'stream': True})
    return download_from_url(*args, **kwargs)

###############################################################################
def download_from_url(url,
                      destination = None,
                      uncompress  = False,
                      user_agent  = None,
                      stream      = False,
                      progress    = False,
                      desc        = None,
                      cleanup     = False,
                      **kwargs):
    """
    Save the resource as a file on disk.
    """
    # Custom user agent if needed #
    if user_agent is not None: magic = make_headers(user_agent)
    else: magic = ""
    # Download #
    if stream: response = request(url, magic, response=True, stream=True, **kwargs)
    else:      content  = request(url, magic, content=True, **kwargs)
    # Get total size #
    if stream:
        total_size = int(response.headers.get('content-length', -1))
        block_size = int(5e5)
    # Sometimes we don't get content-length #
    if stream and total_size < 0:
        return download_from_url(url, destination, uncompress, user_agent,
                                 False, False, **kwargs)
    # Choose the right option for destination #
    destination = handle_destination(url, destination)
    # How the progress bar should look like #
    bar = '{l_bar}{bar}| {n_fmt}/{total_fmt} [{remaining}, ' \
          '{rate_fmt}{postfix}]'
    # Delete the destination if there is any exception or a ctrl-c #
    try:
        # With streaming #
        if stream:
            generator = response.iter_content(chunk_size=block_size)
            if progress:
                # In the future replace with `from tqdm.rich import tqdm`
                from tqdm import tqdm
                bar = tqdm(bar_format   = bar,
                           desc         = desc,
                           total        = total_size,
                           unit         = 'B',
                           unit_scale   = True)
                with open(destination, "wb") as handle:
                    for data in generator:
                        handle.write(data)
                        bar.update(len(data))
                    bar.close()
            else:
                with open(destination, "wb") as handle:
                    for data in generator: handle.write(data)
        # Without streaming #
        if not stream:
            with open(destination, "wb") as handle: handle.write(content)
    # Remove the destination before raising any error #
    except:
        if cleanup and os.path.exists(destination): os.remove(destination)
        raise
    # Uncompress the result #
    if uncompress:
        # To detect tar.gz we rely on the extension #
        if destination.endswith('.tar.gz'):
            return destination.untargz_to()
        # Otherwise, read the magic number #
        with open(destination, 'rb') as f:
            magic = f.read(4)
        # If it's a zip file #
        if magic == b"PK\x03\x04":
            return destination.unzip_to(inplace=True)
        # If it's a gzip file #
        elif magic[:3] == b"\x1f\x8b\x08":
            return destination.ungzip_to()
    # Return #
    return destination

###############################################################################
def handle_destination(url, destination):
    """
    The destination can be either unspecified or can contain either a file path
    or a directory path.
    """
    # Choose a default for destination #
    if destination is None:
        destination = autopaths.tmp_path.new_temp_file()
    # Directory case - choose a filename #
    elif destination.endswith('/'):
        filename    = url.split("/")[-1].split("?")[0]
        destination = Path(destination + filename)
        destination.directory.create_if_not_exists()
    # Normal case #
    else:
        destination = Path(destination)
        destination.directory.create_if_not_exists()
    # Return #
    return destination




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Internal modules #
import inspect, os

# First party modules #
from autopaths import Path

# Constants #
file_name = inspect.getframeinfo(inspect.currentframe()).filename
this_dir  = os.path.dirname(os.path.abspath(file_name)) + '/'
destin    = Path(this_dir + 'user_agent_strings.txt')

###############################################################################
def get_popular_user_agents():
    """
    Retrieve most popular user agent strings.
    Can look at https://techblog.willshouse.com/2012/01/03/most-common-user-agents/
    """
    return ''

###############################################################################
if __name__ == '__main__':
    content = get_popular_user_agents()
    destin.write(content)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os, platform, base64, shutil, gzip
from io import StringIO

# Internal modules #
from plumbing.common import camel_to_snake
from plumbing.cache import property_cached
from plumbing.databases.sqlite_database import SQLiteDatabase

# First party modules #
from autopaths.file_path import FilePath
from autopaths.tmp_path  import new_temp_path

###############################################################################
class AccessDatabase(FilePath):
    """
    A wrapper for a Microsoft Access database via pyodbc.
    On Ubuntu 18 you would install the dependencies like this:

        $ sudo apt install python3-pip
        $ sudo apt install unixodbc-dev
        $ pip install --user pyodbc
    """

    # Enable this to change `ThisName` to `this_name` on all columns #
    convert_col_names_to_snake = False

    # ------------------------------ Constructor ---------------------------- #
    def __init__(self, path,
                       username = 'admin',
                       password = None):
        """
        * The path of the database comes first.
        * The username.
        * The password.
        """
        # Path attribute #
        super(AccessDatabase, self).__init__(path)
        # Attributes #
        self.username = username
        self.password = password
        # Check the database exists #
        self.must_exist()

    # ------------------------------ Properties ----------------------------- #
    @property_cached
    def conn_string(self):
        # Get current system #
        system = platform.system()
        # macOS #
        if system == "Darwin":
            string = "Driver={Microsoft Access Driver (*.mdb)};User Id='%s';DBQ=%s"
            return string % (self.username, self.path)
        # Linux #
        if os.name == "posix":
            string = "Driver={MDBTools};User Id='%s';DBQ=%s"
            return string % (self.username, self.path)
        # Windows #
        if os.name == "nt":
            string = "Driver={Microsoft Access Driver (*.mdb, *.accdb)};User Id='%s';DBQ=%s"
            return string % (self.username, self.path)
        else:
            raise Exception("Unrecognized platform.")

    @property_cached
    def conn(self):
        """To be used externally by the user."""
        return self.new_conn()

    @property_cached
    def own_conn(self):
        """To be used internally in this object."""
        return self.new_conn()

    @property_cached
    def cursor(self):
        """To be used externally by the user."""
        return self.conn.cursor()

    @property_cached
    def own_cursor(self):
        """To be used internally in this object."""
        return self.own_conn.cursor()

    @property
    def tables(self):
        """The complete list of tables."""
        # If we are on unix use mdbtools instead #
        if os.name == "posix":
            import pbs3
            mdb_tables  = pbs3.Command("mdb-tables")
            tables_list = mdb_tables('-1', self.path).split('\n')
            condition   = lambda t: t and not t.startswith('MSys')
            return [t.lower() for t in tables_list if condition(t)]
        # Default case #
        return [table[2].lower() for table in self.own_cursor.tables()
                if not table[2].startswith('MSys')]

    @property
    def real_tables(self):
        """The complete list of tables excluding views and query tables."""
        return [table for table in self.tables if self.test_table(table)]

    # ------------------------------- Methods ------------------------------- #
    def __getitem__(self, key):
        """Called when evaluating ``database[0] or database['P81239A']``."""
        return self.table_as_df(key)

    def __contains__(self, key):
        """Called when evaluating ``'students' in database``."""
        return key.lower() in self.tables

    def test_table(self, table_name):
        """Can the table be read from?"""
        import pyodbc
        try:
            query = "SELECT COUNT (*) FROM `%s`" % table_name.lower()
            self.own_cursor.execute(query)
            self.own_cursor.fetchone()
        except pyodbc.Error:
            return False
        return True

    def new_conn(self):
        """Make a new connection."""
        import pyodbc
        return pyodbc.connect(self.conn_string)

    def close(self):
        self.cursor.close()
        self.conn.close()
        self.own_cursor.close()
        self.own_conn.close()

    def table_must_exist(self, table_name):
        """Return a table as a dataframe."""
        if table_name.lower() not in self.tables:
            raise Exception("The table '%s' does not seem to exist." % table_name)

    def table_as_df(self, table_name):
        """
        Return a table as a dataframe.
        There is a library that can do this, but it has a bug.
        See https://github.com/jbn/pandas_access/issues/3

            import pandas_access
            return pandas_access.read_table(self.path, table_name)

        This is also a possibility https://github.com/gilesc/mdbread
        but it is not in PyPI.
        """
        # Check #
        self.table_must_exist(table_name)
        # If we are on unix use mdb-tools instead #
        if os.name == "posix": df = self.table_as_df_via_mdbtools(table_name)
        # Default case #
        else: df = self.table_as_df_via_query(table_name)
        # Optionally rename columns #
        if self.convert_col_names_to_snake: df = df.rename(columns=camel_to_snake)
        # Return #
        return df

    def table_as_df_via_query(self, table_name):
        """Use an SQL query to create the dataframe."""
        query = "SELECT * FROM `%s`" % table_name.lower()
        import pandas
        return pandas.read_sql(query, self.own_conn)

    def table_as_df_via_mdbtools(self, table_name, *args, **kwargs):
        """Use an mdbtools executable to create the dataframe."""
        import subprocess
        cmd = ['mdb-export', self.path, table_name]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        import pandas
        return pandas.read_csv(proc.stdout, *args, **kwargs)

    def insert_df(self, table_name, df):
        """Create a table and populate it with data from a dataframe."""
        df.to_sql(table_name, con=self.own_conn)

    def count_rows(self, table_name):
        """Return the number of entries in a table by counting them."""
        self.table_must_exist(table_name)
        query = "SELECT COUNT (*) FROM `%s`" % table_name.lower()
        self.own_cursor.execute(query)
        return int(self.own_cursor.fetchone()[0])

    def count_rows_fast(self, table_name):
        """Return the number of entries in a table by using the quick inaccurate method."""
        pass

    def tables_with_counts(self):
        """Return the number of entries in all table."""
        table_to_count = lambda t: self.count_rows(t)
        return zip(self.tables, map(table_to_count, self.tables))

    def drop_table(self, table_name):
        if table_name not in self.tables:
            raise Exception("The table '%s' does not seem to exist." % table_name)
        query = "DROP TABLE %s" % table_name
        self.own_conn.execute(query)

    # ------------------------------- Convert ------------------------------- #
    def convert_to_sqlite(self, destination=None, method="shell", progress=False):
        """Who wants to use Access when you can deal with SQLite databases instead?"""
        # Display progress bar #
        if progress:
            import tqdm
            progress = tqdm.tqdm
        else:
            progress = lambda x:x
        # Default path #
        if destination is None: destination = self.replace_extension('sqlite')
        # Delete if it exists #
        destination.remove()
        # Method with shell and a temp file #
        if method == 'shell':     return self.sqlite_by_shell(destination)
        # Method without a temp file #
        if method == 'object':    return self.sqlite_by_object(destination, progress)
        # Method with dataframe #
        if method == 'dataframe': return self.sqlite_by_df(destination, progress)

    def sqlite_by_shell(self, destination):
        """Method with shell and a temp file. This is hopefully fast."""
        script_path = new_temp_path()
        self.sqlite_dump_shell(script_path)
        from shell_command import shell_output
        shell_output('sqlite3 -bail -init "%s" "%s" .quit' % (script, destination))
        script.remove()

    def sqlite_by_object(self, destination, progress):
        """This is probably not very fast."""
        db = SQLiteDatabase(destination)
        db.create()
        for script in self.sqlite_dump_string(progress): db.cursor.executescript(script)
        db.close()

    def sqlite_by_df(self, destination, progress):
        """Is this fast?"""
        db = SQLiteDatabase(destination)
        db.create()
        for table in progress(self.real_tables): self[table].to_sql(table, con=db.connection)
        db.close()

    def sqlite_dump_shell(self, script_path):
        """Generate a text dump compatible with SQLite by using
        shell commands. Place this script at *script_path*."""
        # First the schema #
        from shell_command import shell_output
        shell_output('mdb-schema "%s" sqlite >> "%s"' % (self.path, script_path))
        # Start a transaction, speeds things up when importing #
        script_path.append("\n\n\nBEGIN TRANSACTION;\n")
        # Then export every table #
        for table in self.tables:
            command = 'mdb-export -I sqlite "%s" "%s" >> "%s"'
            shell_output(command % (self.path, table, script_path))
        # End the transaction
        script_path.append("\n\n\nEND TRANSACTION;\n")

    def sqlite_dump_string(self, progress):
        """Generate a text dump compatible with SQLite.
        By yielding every table one by one as a byte string."""
        # First the schema #
        import pbs3
        mdb_schema = pbs3.Command("mdb-schema")
        yield mdb_schema(self.path, "sqlite").encode('utf8')
        # Start a transaction, speeds things up when importing #
        yield "BEGIN TRANSACTION;\n"
        # Then export every table #
        mdb_export = pbs3.Command("mdb-export")
        for table in progress(self.tables):
            yield mdb_export('-I', 'sqlite', self.path, table).encode('utf8')
        # End the transaction
        yield "END TRANSACTION;\n"

    # --------------------------- Multi-database ---------------------------- #
    def import_table(self, source, table_name):
        """Copy a table from another Access database to this one.
        Requires that you have mdbtools command line executables installed
        in a Windows Subsystem for Linux environment."""
        # Run commands #
        import pbs3
        wsl = pbs3.Command("wsl.exe")
        table_schema   = wsl("-e", "mdb-schema", "-T", table_name, source.wsl_style, "access")
        table_contents = wsl("-e", "mdb-export", "-I", "access", source.wsl_style, table_name)
        # Filter #
        table_schema = ' '.join(l for l in table_schema.split('\n') if not l.startswith("--"))
        # Execute statements #
        self.cursor.execute(str(table_schema))
        self.cursor.execute(str(table_contents))

    # -------------------------------- Create ------------------------------- #
    @classmethod
    def create(cls, destination):
        """Create a new empty MDB at destination."""
        mdb_gz_b64 = """\
        H4sICIenn1gC/25ldzIwMDMubWRiAO2de2wcRx3Hf7O7Pt/d3u6eLyEtVaOaqg+EkjQvuVVDwa9a
        jWXHdZxQQlCJ7fOrfp3OTpqkhVxTItFWIhVQVFBRVNIKRaColVpAUKGKRwwFqUAhKiBIpUaoVWP+
        qKgIIHL8Znb39u72znWJiWP3+9l473fzm/nNY3cdf2fmbBJEPdO9E+nebLq+fWC6vrWZOImen9D7
        9sR+vPPNE0PZxo/TE5879mj+yNc3/OzAD2bXv3DmV9/o/8PZnxxr+/fDL2w79ulzN7e+/sS/zvzz
        w3+N1z28p3PTfQ3nfn/m2YmeFS2no89uWnvqwO5HUvd/5Phr938tes3j/zm5+qT41J8/P/iZx87/
        +qHrjgyduubG1t/+7eWB2XztTNuT+1clZt9c2/e7HRGizevWEwAAAAAAAACAhUEIwvE+PoRIO8K7
        FzT6obPPwTMBAAAAAAAAAABcfpzPXwya+Ispo1xlEO2KEEX9eaGyWnrqyKQ60tQ0AcNZRcR1RYuy
        +XZCxoqRzmaMI6cKGRJuJVrIEZUOQ9UrHStUYpyzKkdNmSPFDkM6aguhXMdVHCMuHXE2Suu4IFQJ
        l6CErNWUDouDlbdKOZIcrKLD4S5WdNhqIEodqlVaofKgVTHpiBQ6uLG0uaKsuYbf3IS8BmV1qFAm
        j1Z5Hbp06GWDKC+DTS00SRN8DFA/TXNfW6mXX3upj7+mOHWllzLAObN8du0gdSdlKO3ZcWqjMbaH
        uOQqtidViRF+P0HbOH2c3xm0lfMb1EH7uHZ5vp32c+ks+5PqfSeXS9NejjTAvZQpd7J3kuuJFqLE
        qYvuVa3Ocqk7OVXWNMFxZPRVtJ1zSXuCBrlkh+rjEF1Zlt5Dw6qN0xx5Bx3gGgbowVo56EIjkc9T
        xX9Jdd+5PKDOD6q3VQvwv7qiZ8st419cdYHlo6iuriF8X4HA590AsodXhvrsj0yMDPnAuI+ZvOrq
        1o7K51Hdy7a8cdXNm5AedbfG5W3j3lOybxFZKb6zAgAAAAAAsNzQxAlbvnYJV3VcUU3/S2luBIKF
        ha+IlWp+wxW4IiRXRSXxKeNU1eOxUuUbSOIINbEM7WT506ZE3LASgCOeYJWCMcnCsI/u8eSsFEYR
        lnlbWa6+u0jTYqSkvuQL9G5CLFwTRBMAAAAAAAAAgMtW/79lyVdLKxW7oqDF3bXOniib0UD/m/xq
        loWqvFwt3DX/mrLNALIu3V35NkpK1JDmL+2XOmr9pf1gKiFY4I672wc0mveaf6zaenyKmljPT6t5
        hT7a6y13y0XqjFpwneJjRC0oRwvL3eUL2fHCcuyGIntjhTkDuZCd5Vc5j+HNUMyx+myYcpHW5YG5
        ZijUdbg2VFu4ZzzcHFM3seQLAAAAAAAAAMtc//9S6cm1emX97ytK1v81rHelhtfVfAFnseZXRdV9
        Ad7+dhGS5kbl3eqe/K8pU/nnYwX5X2VeoLbCZwHi7txD6aTELabnoLJ5AfPFC8JmFd3Pun+MlfM4
        q/846/4s62i5+8Dmc7EvSVN0UG2tL00p1uPXqZTt/G5QqX+5lbufz+mSctVzFce6upBrTG3Fd+cn
        pmiYrUyw8+GNfL4hn8/k83qZrVlyGzgPeqbhjcOqx7KMEZRpU/MPQ+rsldEtuYm8vExkznoMS+6b
        KC5TZRt8wVf4xEkFX4V5D/X2vYz1/EcR8yMAAAAAAACAJY0Qf/d3vLPUlb//b4Nzzv6W3Wevtl+1
        vmxts2LWTxOHErcm3jGfMUfNG0yMGQAAAAAAeJ/8rLwAMXIYRgCARFv8IIaYtKpGqCdqlN/2kupD
        /ob67qXhsi0lDh2Vp6728faO9tHuUflfWJ1wE0e6724f35XuG71r16Dr0FwH573by6rKi0N7RveN
        tnd6aTVBWrpjd3fnuJtsBMnDk90ju7zckSA5XGGtdGrK2dWhUnRcMgAAAAAAAAD4v2CIV6vqf82I
        Jusbcwsy7wkWSf/n1JQNq/Oc+uQGq/ecmsphYZ6Tn6XwRLjwxb7mTxDoakLgURUFshwAAAAAAAAA
        ljpCrHZ8W/f2/2NUAAAAAAAAAAAAhXH5RLm4IIbotqot7hbW/0MGWCp46/+pgpHwjZS3IyAlfMPy
        tgakNN+wfcPxNgukdN9I+kadt30gZfhGjW+s8I2V3s6CVNTbWZCK+Eatb3zAN1Z5mw5SMd+I+wZ+
        +QQAAAAAAAAA/K8IcdT27Zqi3/+HkQEAAAAAAAAAsGgkMQQLjSHqbQPDAAAAAAAAAAAALGuw/g8A
        AAAAAAAA4DJUqwsQI7cQDWlcLiMq1/9rcGMBAAAAAAAAAADLGuh/AAAAAAAAAAAA+h8AAAAAAAAA
        AABLHyHusDTPjtLzTtoxnRftUftqe8YatDA+AAAAAAAAAPDeqJN/KVt+et0R9PYnzz7W8PrZRv+V
        HblO6qEDNEXbaYDGqJemaYQmaYJThtnK8Gvzb1opfDRTPZmUlxUY86qgm/ZyFVkOOqCC3kLhoyEI
        qs8raBO10O0q3EYKH+uDcNq8wnVRH93D7evnYZhHG5kkB3a0OYO2ctCWV9ZR+FhT0l2HCzl6xVBz
        XZyPUvi4taTjcwRuVUF7uYW9HMy9MJspfGwMAoo5A+5Qwca8UHN2WogeU/fu0ito1vmjM+M85zzp
        fNG5zxl2djrNzk3O9+0m+yWrx2q0fpH4buJ4Yk3ig4lvmkfxx9gBAAAAAAC4OAylQfJ5h5pfSVCc
        f853gqSmWPSZux6xjUznltH2HT/flNu7++0NZ7/07cg/vnPbVu30y6d/NLvlabPh+j81v/Xc5g9l
        1h2f+epn9+VPdN90OHHvU50fm94y/ZXvWQ/tP/yJG/NH3llz8A79tlNPG72DHSePHdzz2s3XPzVj
        vzSUvSHjVys1Rv5CSUv8pEvcEqkbV/KX35JaQ+npikmRS9o4rtYIt8RYnJa4Ou6SV6stTm+l7rcX
        q9qSy+23pCVIcgV/SZKuJj5CSRc4Y/PpkiesLJcI53J37NvFuQzv4peGL0/SypP+C+45xVAAMAEA
        """
        pristine = StringIO()
        pristine.write(base64.b64decode(mdb_gz_b64))
        pristine.seek(0)
        pristine = gzip.GzipFile(fileobj=pristine, mode='rb')
        with open(destination, 'wb') as handle: shutil.copyfileobj(pristine, handle)
        return cls(destination)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Third party modules #
import brewer2mpl

###############################################################################
# Nice colors from the brewer project #
colors = brewer2mpl.get_map('Set1', 'qualitative', 8).mpl_colors
colors.reverse()
colors += brewer2mpl.get_map('Set2',    'qualitative', 8).mpl_colors
colors += brewer2mpl.get_map('Set3',    'qualitative', 8).mpl_colors
colors += brewer2mpl.get_map('Pastel1', 'qualitative', 8).mpl_colors
colors += brewer2mpl.get_map('Pastel2', 'qualitative', 8).mpl_colors
colors += brewer2mpl.get_map('Greys',   'sequential',  8).mpl_colors

###############################################################################
# Nice colors for plotting stacked bar charts #
other_colors = ["#9b59b6", "#3498db", "#95a5a6", "#e74c3c", "#34495e", "#2ecc71",
                "#FCD124", "#5F0061", "#FBBEBA", "#AAC2F6", "#606925", "#469A55",
                "#ABD8B2", "#B4B1ED", "#613162", "#76C3E5", "#F05575", "#48CBD1"]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in module #
import time

# Internal modules #
from plumbing.cache import property_cached
from plumbing.common import pretty_now

###############################################################################
class Employee(object):

    @property_cached
    def check_in_time(self):
        print("Evaluating...")
        return pretty_now()

###############################################################################
bob = Employee()
print(bob.check_in_time)

time.sleep(2)
print(bob.check_in_time)

bob.__dict__.get('__cache__').pop('check_in_time')
print(bob.check_in_time)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

###############################################################################
class MultiDownload:
    """
    Taken and adapted from:
    https://github.com/soda480/pypbars/blob/main/examples/example3.py

    But should take a look at:
    https://pypi.org/project/tqdm-multiprocess/
    """

    # The number of files to download in parallel #
    num_proc = 3

    def prepare_queue(self, queue):
        for _ in range(20):
            import random
            queue.put({'total': random.randint(100, 150)})

    def download_one(self, total):
        for index in range(total):
            import time, random
            time.sleep(random.choice([.001, .003, .008]))
        return total

    def run_queue(self, worker_id, queue):
        result = 0
        while True:
            try:
                total = queue.get(timeout=1)['total']
                result += self.download_one(worker_id)
            except Empty:
                break
        return result

    def download_parallel(self):
        # Imports #
        from multiprocessing import Queue, Pool, get_context
        from pypbars import ProgressBars
        from list2term.multiprocessing import LinesQueue
        from list2term.multiprocessing import QueueManager
        # Make two queues #
        QueueManager.register('LinesQueue', LinesQueue)
        QueueManager.register('Queue', Queue)
        # Enter the queue #
        with QueueManager() as manager:
            # The lines queue #
            lines_queue = manager.LinesQueue(ctx=get_context())
            # The data queue #
            data_queue = manager.Queue()
            # Add all jobs to the queue #
            self.prepare_queue(data_queue)
            # Start a pool of workers #
            with Pool(self.num_proc) as pool:
                # Arguments that will be sent to the function #
                process_data = [('name%i' % i, data_queue, lines_queue)
                                for i in range(self.num_proc)]
                # Map our function to each item #
                results = pool.starmap_async(self.run_queue, process_data)
                # Get the lookup #
                lookup = [f'{data[0]}' for data in process_data]
                # Create ProgressBars #
                with ProgressBars(lookup        = lookup,
                                  show_prefix   = False,
                                  show_fraction = False,
                                  use_color     = True,
                                  show_duration = True,
                                  clear_alias   = True) as lines:
                    while True:
                        try:
                            item = lines_queue.get(timeout=0.1)
                            if item.endswith('->reset'):
                                i, message = lines.get_index_message(item)
                                lines[i].reset(clear_alias=False)
                            else:
                                lines.write(item)
                        except Empty:
                            if results.ready():
                                for i, _ in enumerate(lines):
                                    lines[i].complete = True
                                break
        # Return #
        return results.get()

###############################################################################
if __name__ == '__main__':
    print(MultiDownload().download_parallel())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# First party modules #
from autopaths import Path

# Internal modules #
from plumbing.cache import property_cached

###############################################################################
class MultiDataFrameXLS:
    """
    Takes several dataframes and writes them to an XLS file.
    The dataframes are spread through different work sheets.

    In addition, each work sheet can contain an arbitrary number
    of dataframes.

    You have to provide a dictionary where:

    * Each key is the name of a given work sheet in the final XLS as a string.
    * Each value is a list containing an arbitrary number of dictionaries.

    * Each one of these dictionaries must contain a DataFrame in the
      'dataframe' key, as well as optional extra labels as seen below.

            sheet = {
                'dataframe': df,
                'title':     "Best dataframe ever",
                'x_title':   "Foot length",
                'y_title':   "Yearly income",
                'x_label':   None,
                'y_label':   None,
                'x_extra':   None,
                'y_extra':   None,
            }
    """

    # Parameters #
    spacing = 6
    indentation = 1

    def __init__(self, sheet_to_dfs, path):
        self.sheet_to_dfs = sheet_to_dfs
        self.path         = Path(path)

    def __call__(self):
        """
        Write several dataframes, to several excel sheets.
        """
        # Imports #
        import pandas
        # Create path if not exists #
        self.path.directory.create_if_not_exists()
        # Create a writer #
        self.writer = pandas.ExcelWriter(str(self.path), engine='xlsxwriter')
        # Create a sheet per every key #
        for key in self.sheet_to_dfs:
            worksheet = self.writer.book.add_worksheet(key)
            self.writer.sheets[key] = worksheet
        # Write each sheet #
        for key in self.sheet_to_dfs: self.write_one_sheet(key)
        # Save #
        self.writer.save()
        # Return #
        return self.path

    def write_one_sheet(self, key):
        """
        Write several dataframes, all to the same excel sheet.
        It will append a custom title before hand for each
        dataframe.
        """
        # Imports #
        import pandas
        # Get sheet #
        sheet = self.writer.sheets[key]
        # Get dataframes #
        all_dfs = self.sheet_to_dfs[key]
        # Initialize #
        row = 0
        # Loop #
        for info in all_dfs:
            # Get dataframe #
            df = info['dataframe']
            # Write custom title #
            sheet.write_string(row, 0, info.get('title', ''))
            row += 2
            # Add extras #
            df.index.name   = info.get('y_extra', '')
            df.columns.name = info.get('x_extra', '')
            # Add Y labels #
            title, label = info.get('y_title', ''), info.get('y_label', '')
            df = pandas.concat({title: df}, names=[label])
            # Add X labels #
            title, label = info.get('x_title', ''), info.get('x_label', '')
            df = pandas.concat({title: df}, names=[label], axis=1)
            # Write dataframe #
            df.to_excel(self.writer,
                        sheet_name = key,
                        startrow   = row,
                        startcol   = self.indentation)
            # Increment #
            row += len(df.index) + self.spacing

###############################################################################
class ConvertExcelToCSV:
    """
    Will convert an excel file into its CSV equivalent.
    Can support multiple work sheets into a single CSV.
    """

    def __init__(self, source_path, dest_path, **kwargs):
        # Record attributes #
        self.source = Path(source_path)
        self.dest   = Path(dest_path)
        # Keep the kwargs too #
        self.kwargs = kwargs
        # Check directory case #
        if self.dest.endswith('/'):
            self.dest = self.dest + self.source.filename
            self.dest = self.dest.replace_extension('csv')

    def __call__(self):
        """Are we mono or multi sheet?"""
        if len(self.handle.sheet_names) > 1: self.multi_sheet()
        else:                                self.mono_sheet()

    @property_cached
    def handle(self):
        """Pandas handle to the excel file."""
        import pandas
        return pandas.ExcelFile(str(self.source))

    def mono_sheet(self):
        """Supports only one work sheet per file."""
        import pandas
        xls = pandas.read_excel(str(self.source))
        xls.to_csv(str(self.dest), **self.kwargs)

    def multi_sheet(self):
        """
        Supports multiple work sheets per file.
        Will concatenate sheets together by adding an extra column
        containing the original sheet name.
        """
        # Imports #
        import pandas
        # Initialize #
        all_sheets = []
        # Loop #
        for name in self.handle.sheet_names:
            sheet = self.handle.parse(name)
            sheet.insert(0, "nace", name)
            all_sheets.append(sheet)
        # Write #
        df = pandas.concat(all_sheets)
        df.to_csv(str(self.dest), **self.kwargs)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
from datetime import datetime

# Internal modules #
from plumbing.cache import property_cached
from plumbing.color import Color

###############################################################################
class Timer(object):
    """
    Useful for timing the different steps in a pipeline.

    Use it like this:

        from plumbing.timer import Timer
        timer = Timer()
        timer.print_start()
        timer.print_elapsed()
        timer.print_end()
        timer.print_total_elapsed()

    Or like this:

        with Timer():
            time.sleep(2)
    """

    format = "%Y-%m-%d %H:%M:%S %Z%z"

    def get_now(self):
        return datetime.now(self.time_zone)

    def to_str(self, time):
        return time.strftime(self.format)

    def __init__(self, msg=None):
        import tzlocal
        self.time_zone  = tzlocal.get_localzone()
        self.start_time = self.get_now()
        self.last_mark  = self.get_now()
        self.msg        = msg

    def __enter__(self):
        """Start the timer and print."""
        self.start_time = self.get_now()
        self.print_start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Stop the timer and print."""
        self.print_end()
        self.print_total_elapsed(self.msg)

    def print_start(self, msg=""):
        # Parameters #
        time = self.start_time
        # Make message #
        msg += "Start at: %s" % self.to_str(time)
        # Print #
        self.print(self.prefix + msg + self.suffix)

    def print_end(self, msg=""):
        # Set #
        self.end_time = self.get_now()
        # Parameters #
        time = self.end_time
        # Make message #
        msg += "End at: %s" % self.to_str(time)
        # Print #
        self.print(self.prefix + msg + self.suffix)

    def print_elapsed(self, reset=True, msg=""):
        # Parameters #
        time = self.get_now() - self.last_mark
        # Make message #
        msg += " - Elapsed time: %s" % time
        # Print #
        self.print(self.prefix + msg + self.suffix)
        # Set #
        if reset: self.last_mark = self.get_now()

    def print_total_elapsed(self, reset=True, msg=""):
        # Parameters #
        time = self.get_now() - self.start_time
        # Make message #
        msg += "Total elapsed time: %s"
        msg = msg % time
        # Print #
        self.print(self.prefix + msg + self.suffix)
        # Set #
        if reset: self.last_mark = self.get_now()

    @property_cached
    def color(self):
        """
        Should we use color or not ?
        If we are not in a shell like ipython then not.
        """
        import __main__ as main
        return not hasattr(main, '__file__')

    @property
    def prefix(self): return "" if not self.color else Color.grn

    @property
    def suffix(self): return "" if not self.color else Color.end

    def print(self, *args, **kwargs):
        print(*args, **kwargs)

###############################################################################
class LogTimer(Timer):
    """
    Same as a Timer, except the messages go to a logger object instead of
    being printed to the standard out.
    """

    def __init__(self, log, high_precision=False):
        super(LogTimer, self).__init__(high_precision)
        self.color = False
        self.log = log

    def print(self, msg):
        self.log.info(msg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# First party modules #
from autopaths.file_path import FilePath

###############################################################################
def check_blocked_request(tree):
    """
    Check if the server denied the request.
    And raise an exception if it was.
    """
    # Modules #
    from lxml import etree
    # Did we get a filepath? #
    if isinstance(tree, FilePath):
        if tree.count_bytes > 1000000: return
        tree = tree.contents
    # Did we get a tree or raw text? #
    if isinstance(tree, str): tree = etree.HTML(tree)
    # By default we are good #
    blocked = False
    # Try Incapsula #
    blocked = blocked or check_incapsula(tree)
    # If we were indeed blocked, we can stop here #
    if blocked: raise Exception("The request was flagged and blocked by the "
                                "server.")

###############################################################################
def check_incapsula(tree):
    # By default, we are good #
    blocked = False
    # Result type 1 from Incapsula #
    meta = tree.xpath("//head/meta[@name='ROBOTS']")
    if meta and 'NOINDEX' in meta[0].get('content'): blocked = True
    # Result type 2 from Incapsula #
    meta = tree.xpath("//head/meta[@name='robots']")
    if meta and 'noindex' in meta[0].get('content'): blocked = True
    # If we were indeed blocked, we can stop here #
    return blocked


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

###############################################################################
def flexible_join(first, other, on=None, how=None, lsuffix=None, rsuffix=None):
    """
    Implement a common join pattern with pandas set_index()
    on both data frames followed by a reset_index() at the end.

    TODO check the data types (dtypes) of all the index columns
     are matching on both sides of the join.
    """
    # Check if `on` is a set #
    if isinstance(on, set): on = list(on)
    # Set indexes #
    if on is not None:
        first  = first.set_index(on)
        other  = other.set_index(on)
    # Join #
    result = first.join(other, how=how, lsuffix=lsuffix, rsuffix=rsuffix)
    # Reset index #
    result = result.reset_index()
    # Return #
    return result

###############################################################################
def left_join(*args, **kwargs):
    """This method is patched onto pandas.DataFrame for convenience."""
    return flexible_join(*args, **kwargs, how='left')

def right_join(*args, **kwargs):
    """This method is patched onto pandas.DataFrame for convenience."""
    return flexible_join(*args, **kwargs, how='right')

def inner_join(*args, **kwargs):
    """This method is patched onto pandas.DataFrame for convenience."""
    return flexible_join(*args, **kwargs, how='inner')

def outer_join(*args, **kwargs):
    """This method is patched onto pandas.DataFrame for convenience."""
    return flexible_join(*args, **kwargs, how='outer')

###############################################################################
def patch_pandas():
    """Add a nice method to pandas.DataFrame objects."""
    import pandas
    pandas.DataFrame.left_join  = left_join
    pandas.DataFrame.right_join = right_join
    pandas.DataFrame.inner_join = inner_join
    pandas.DataFrame.outer_join = outer_join


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio

Script to test the parsing of a `setup.py`file.
"""

# Built-in modules #
import os

# Internal modules #
from plumbing.dependencies import check_setup_py

###############################################################################
path = os.path.expanduser("~/repos/pacmill/setup.py")
result = check_setup_py(path)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# First party modules #
from plumbing.graphs import Graph
from plumbing.cache import property_cached

###############################################################################
class Multiplot(Graph):
    """
    Similar to a FacetPlot, expect we don't use `seaborn` and do everything
    ourselves with `matplotlib`.
    """

    # Size of grid #
    n_rows = 1
    n_cols = 1

    # Defaults #
    share_y = True
    share_x = True

    # Defaults #
    height = None
    width  = None

    @property_cached
    def fig_and_axes(self):
        # Synonyms #
        if self.share_x is True: self.share_x = 'all'
        if self.share_y is True: self.share_y = 'all'
        # Figure size #
        width  = self.n_cols * 5 if self.width is None else self.width
        height = self.n_rows * 5 if self.height is None else self.height
        # Create #
        from matplotlib import pyplot
        fig, axes = pyplot.subplots(nrows   = self.n_rows,
                                    ncols   = self.n_cols,
                                    sharex  = self.share_x,
                                    sharey  = self.share_y,
                                    figsize = (width, height))
        # Return #
        return fig, axes

    @property
    def fig(self): return self.fig_and_axes[0]

    @property
    def axes(self): return self.fig_and_axes[1]

    #--------------------------- Convenience ---------------------------------#
    def iterate_all_axes(self, fn):
        import numpy
        for axes in numpy.nditer(self.axes, flags=['refs_ok']): fn(axes[()])

    def x_grid_on(self):
        """Add horizontal lines on the x axis."""
        fn = lambda axes: axes.xaxis.grid(True, linestyle=':')
        self.iterate_all_axes(fn)

    def y_grid_on(self):
        """Add horizontal lines on the y axis."""
        fn = lambda axes: axes.yaxis.grid(True, linestyle=':')
        self.iterate_all_axes(fn)

    def set_x_lim(self, a, b):
        fn = lambda axes: axes.set_xlim(a, b)
        self.iterate_all_axes(fn)

    def set_y_lim(self, a, b):
        fn = lambda axes: axes.set_ylim(a, b)
        self.iterate_all_axes(fn)

    def set_x_ticks(self, ticks):
        fn = lambda axes: axes.set_xticks(ticks)
        self.iterate_all_axes(fn)

    def set_y_ticks(self, ticks):
        fn = lambda axes: axes.set_yticks(ticks)
        self.iterate_all_axes(fn)

    def set_x_tick_labels(self, labels):
        fn = lambda axes: axes.set_xticklabels(labels)
        self.iterate_all_axes(fn)

    def set_y_tick_labels(self, labels):
        fn = lambda axes: axes.set_yticklabels(labels)
        self.iterate_all_axes(fn)

    def hide_titles(self):
        """Remove the subplot titles."""
        fn = lambda axes: axes.title.set_visible(False)
        self.iterate_all_axes(fn)

    def y_max_two_decimals(self):
        """Force maximum two decimals for y axis."""
        import matplotlib
        str_formatter = matplotlib.ticker.FormatStrFormatter('%.2f')
        fn = lambda axes: axes.yaxis.set_major_formatter(str_formatter)
        self.iterate_all_axes(fn)

    def set_x_labels(self, label):
        fn = lambda axes: axes.set_xlabel(label)
        self.iterate_all_axes(fn)

    def set_y_labels(self, label):
        fn = lambda axes: axes.set_ylabel(label)
        self.iterate_all_axes(fn)

    def remove_frame(self):
        fn = lambda axes: axes.spines["top"].set_visible(False)
        self.iterate_all_axes(fn)
        fn = lambda axes: axes.spines["right"].set_visible(False)
        self.iterate_all_axes(fn)

    def hide_full_axes(self, axes):
        axes.xaxis.grid(False)
        axes.yaxis.grid(False)
        axes.spines["top"].set_visible(False)
        axes.spines["right"].set_visible(False)
        axes.spines["left"].set_visible(False)
        axes.spines["bottom"].set_visible(False)
        axes.set_xlabel(None)
        # X axis #
        for xlabel in axes.get_xticklabels(): xlabel.set_visible(False)
        for tick in axes.get_xticklines(): tick.set_visible(False)
        # Y axis #
        for ylabel in axes.get_yticklabels(): ylabel.set_visible(False)
        for tick in axes.get_yticklines(): tick.set_visible(False)

    def y_center_origin(self):
        """
        Place the zero-intercept exactly the middle of the graph.
        This is not the same as doing:
        `axes.spines['left'].set_position('center')`
        """
        def fn(axes):
            bottom, top = axes.get_ylim()
            highest     = max(abs(top), abs(bottom))
            axes.set_ylim(-highest, highest)
        self.iterate_all_axes(fn)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair.
MIT Licensed.
Contact at www.sinclair.bio
"""

# Built-in modules #
import os, time, inspect, tempfile, pickle, hashlib, base64

# First party modules #
from autopaths import Path
from autopaths.dir_path import DirectoryPath
from autopaths.file_path import FilePath

###############################################################################
def cached(f):
    """Decorator for functions evaluated only once."""
    def memoized(*args, **kwargs):
        if hasattr(memoized, '__cache__'):
            return memoized.__cache__
        result = f(*args, **kwargs)
        memoized.__cache__ = result
        return result
    return memoized

###############################################################################
class property_cached(object):
    """
    This decorator converts a method with a single self argument
    into a property cached on the instance.

    In other words, it's like `@property` but memoized.

    Since Python 3.6 we have `@functools.cached_property` which is
    mostly the same as this.

        >>> from plumbing.cache import property_cached
        >>>
        >>> class Square:
        >>>     def __init__(self, size):
        >>>         self.size = size
        >>>
        >>>     @property_cached
        >>>     def area(self):
        >>>         print("Evaluating...")
        >>>         return self.size * self.size
        >>>
        >>> shape = Square(5)
        >>> print(shape.size)
        >>> print(shape.area)  # Area computed the first time
        >>> shape.area = 99    # Now update an attribute
        >>> print(shape.area)  # Same result returned, not recomputed
        >>> del shape.area     # This will purge the cache
        >>> print(shape.area)  # Updated result, it's computed again
    """

    def __init__(self, func):
        """
        If you see this docstring or code in a seemingly unrelated part of
        an auto-generated documentation, it means the program making the
        documentation was unable to correctly traverse a decorated property.
        """
        # Record the function that we are decorating #
        self.func = func
        # Set the documentation string of the underlying function here #
        self.__doc__ = func.__doc__
        self.__get__.__func__.__doc__ = func.__doc__
        self.__init__.__func__.__doc__ = func.__doc__
        # Get the plain name of the function #
        self.name = self.func.__name__

    def __get__(self, instance, owner):
        """
        If you see the current source code in a seemingly unrelated part of
        an auto-generated documentation, it means the program making the
        documentation was unable to correctly traverse a decorated property.
        """
        # For debugging purposes #
        if False: print("-> property cached `%s`" % instance)
        # If called from a class #
        if instance is None: return self
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Is the answer in the cache? #
        if self.name in instance.__cache__: return instance.__cache__[self.name]
        # If not we will compute it #
        if inspect.isgeneratorfunction(self.func): result = tuple(self.func(instance))
        else:                                      result = self.func(instance)
        # Let's store the answer for later #
        instance.__cache__[self.name] = result
        # Return #
        return result

    def __set__(self, instance, value):
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Overwrite the value #
        instance.__cache__[self.name] = value

    def __delete__(self, instance):
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Remove the key #
        instance.__cache__.pop(self.name, None)

    def check_cache(self, instance):
        if '__cache__' not in instance.__dict__: instance.__cache__ = {}

###############################################################################
class property_pickled:
    """
    Same thing as `property_cached` but the cache will be stored on disk
    with the `pickle` module. So you should check that the return value of the
    function you decorate can be pickled. Otherwise, checkout the `dill`
    module.

    The path of the pickle file will be determined by looking for the
    `cache_dir` attribute of the instance containing the cached property
    and combining the function name with a '.pickle' at the end.

    If no `cache_dir` attribute exists it, a default location will be
    chosen (tmpdir). But this will have for effect that all instances of the
    class will have the same cached value (works well for singletons only).

    The location will default to the temporary directory plus a very short
    hash of the function's import path, so you could get something like this:

        '/var/temporary/pickled_properties/EfEZTAubgXI'
    """

    def __init__(self, func, at=None, path=None):
        """
        If you see this docstring or code in a seemingly unrelated part of
        an auto-generated documentation, it means the program making the
        documentation was unable to correctly traverse a decorated property.
        """
        # Record the function that we are decorating #
        self.func = func
        # Set the documentation string of the underlying function here #
        self.__doc__ = func.__doc__
        self.__get__.__func__.__doc__ = func.__doc__
        self.__init__.__func__.__doc__ = func.__doc__
        # Get the plain name of the function #
        self.name  = self.func.__name__
        # Optionally, specify the name of the property that will tell us the
        # path at which we should pickle.
        self.at = at
        # Optionally, specify the path at which we should pickle directly #
        self.path = path

    def __get__(self, instance, owner, debug=False):
        """
        If you see this docstring or code in a seemingly unrelated part of
        an auto-generated documentation, it means the program making the
        documentation was unable to correctly traverse a decorated property.
        """
        # If called from a class #
        if instance is None: return self
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Is the answer in the cache? #
        if self.name in instance.__cache__: return instance.__cache__[self.name]
        # Where should we look in the file system ? #
        path = self.get_pickle_path(instance)
        path.directory.create_if_not_exists()
        # Debugging message #
        if debug:
            msg = "-> The pickle cache path for '%s' is '%s'."
            print(msg % (instance, path))
        # Is the answer already on the file system? #
        if path.exists:
            with open(path, 'rb') as handle: result = pickle.load(handle)
            instance.__cache__[self.name] = result
            return result
        # If not we will compute it #
        if inspect.isgeneratorfunction(self.func): result = tuple(self.func(instance))
        else:                                      result = self.func(instance)
        # Let's store the answer for later in the cache #
        instance.__cache__[self.name] = result
        # And also store it on the disk #
        with open(path, 'wb') as handle: pickle.dump(result, handle)
        # Return #
        return result

    def __set__(self, instance, value):
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Overwrite the value in memory #
        instance.__cache__[self.name] = value
        # Where should we look in the file system ? #
        path = self.get_pickle_path(instance)
        path.directory.create_if_not_exists()
        # And also overwrite it on the disk #
        with open(path, 'wb') as handle: pickle.dump(value, handle)

    def __delete__(self, instance):
        # Does a cache exist for this instance? #
        self.check_cache(instance)
        # Remove the key #
        instance.__cache__.pop(self.name, None)
        # And remove the file on disk #
        path = self.get_pickle_path(instance)
        path.remove()

    def check_cache(self, instance):
        if '__cache__' not in instance.__dict__: instance.__cache__ = {}

    def get_pickle_path(self, instance):
        # First check if an `at` parameter was specified #
        if self.at is not None:
            return Path(getattr(instance, self.at))
        # Secondly check if a `path` parameter was specified #
        if self.path is not None:
            if isinstance(self.path, str):      self.path = Path(self.path)
            if isinstance(self.path, FilePath): return self.path
            # If it's a directory we can create the full file path #
            if isinstance(self.path, DirectoryPath):
                if isinstance(instance, FilePath):
                    return instance.replace_extension(self.name + '.pickle')
                else:
                    return self.path + self.name + ".pickle"
        # Thirdly check if the instance has a cache_dir specified #
        if 'cache_dir' in instance.__dict__:
            return Path(instance.cache_dir + self.name + '.pickle')
        # Fourthly, check if the instance has autopaths activated and if
        # there is a cache directory specified there.
        if hasattr(instance, 'autopaths'):
            directory = instance.autopaths.get('cache_dir')
            if directory is not None:
                return directory + self.name + '.pickle'
        # Otherwise, we go the default route (no instance passed) #
        return self.get_default_path()

    def get_default_path(self):
        # Get the temporary directory (platform dependant) #
        path = tempfile.gettempdir() + '/pickled_properties/'
        # Create that directory if it doesn't exist #
        os.makedirs(path, exist_ok=True)
        # Find the function's import path in the package namespace #
        func_loc = self.func.__module__ + '.' + self.name
        # Make a short safe name from the import path #
        short_name = hashlib.md5(func_loc.encode()).digest()[:8]
        short_name = base64.urlsafe_b64encode(short_name).decode()
        short_name = short_name.replace('=','')
        # Return #
        return Path(path + short_name)

###############################################################################
def property_pickled_at(at):
    """
    Same thing as above, but you can specify the name of another property as
    a string, that will be called on the instance once to determine the
    path at which to write and load the pickle file from. This property should
    hence return the same path for every equivalent instance.
    """
    def wrapper(function): return property_pickled(function, at=at)
    return wrapper

###############################################################################
def property_pickled_path(path):
    """
    Same thing as above, but you can specify the path of the pickle file
    directly in the decorator (instead of specifying a property name).
    """
    def wrapper(function): return property_pickled(function, path=path)
    return wrapper

###############################################################################
def expiry_every(seconds=0):
    def memoize_with_expiry(func, *args, **kwargs):
        # Get the cache #
        if not hasattr(func, '__cache__'): func.__cache__ = [(0,0)]
        cache = func.__cache__
        # Check the cache #
        if cache:
            result, timestamp = cache[0]
            age = time.time() - timestamp
            if age < seconds: return result
        # Update the cache #
        result = func(*args, **kwargs)
        cache[0] = (result, time.time())
        # Return #
        return result
    from decorator import decorator
    return decorator(memoize_with_expiry)

###############################################################################
class LazyString(object):
    """A string-like object that will only compute its value once, when accessed."""

    def __str__(self): return self.value

    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value is None: self._value = self.function()
        return self._value

###############################################################################
class LazyDict(object):
    """A dictionary-like object that will only compute its value once, when accessed."""

    def __getitem__(self, item): return self.value[item]

    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value is None: self._value = self.function()
        return self._value

###############################################################################
class LazyList(object):
    """A list-like object that will only compute its value once, when accessed."""

    def __iter__(self): return iter(self.value)
    def __len__(self):  return len(self.value)

    def __init__(self, function):
        self._value = None
        self.function = function

    @property
    def value(self):
        if self._value is None: self._value = self.function()
        return self._value

###############################################################################
class class_property(property):
    """
    You can use this like this:

        class Foo(object):
            @class_property
            @classmethod
            @cached
            def bar(cls):
                return 1+1
    """

    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

###############################################################################
class invalidate_cache(object):
    """
    Adds a hook to allow for another property setter to
    invalidated a part of the cache.

        class Square(object):
            def __init__(self, size):
                self._size = size

            @property
            def size(self):
                return self._size

            @cached_property
            def area(self):
                print("Evaluating...")
                return self.size * self.size

            @size.setter
            @invalidate_cache("area")
            def size(self, size):
                self._size = size

    shape = Square(5)
    print(shape.size)
    print(shape.area)
    shape.size = 6
    print(shape.area)
    """

    def __init__(self, func):
        self.func    = func
        self.__doc__ = getattr(func, '__doc__')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to test the functionality of the property_pickled decorator.

Written by Lucas Sinclair. MIT Licensed.

Call it like this:

    $ ipython3 -i ~/repos/plumbing/test/cache/property_pickled_at.py
"""

# Built-in modules #
import os

# Internal modules #
from plumbing.cache import property_pickled_at

###############################################################################
class Square:
    def __init__(self, size):
        self.size = size

    @property_pickled_at("area_cache_path")
    def area(self):
        print("Evaluating...")
        return self.size * self.size

    @property
    def area_cache_path(self):
        path = "~/repos/plumbing/tests/cache/area_cache.pickle"
        return os.path.expanduser(path)

###############################################################################
shape = Square(5)
print(shape.area)